/* Copyright (c) 2012 Dovecot Oy, see the included COPYING file */

#include "lib.h"
#include "buffer.h"
#include "unichar.h"
#include "mmap-util.h"
#include "mail-user.h"
#include "mail-storage-hooks.h"
#include "icu-plugin.h"

#include <sys/stat.h>
#include <unicode/udata.h>
#include <unicode/ustring.h>
#include <unicode/unorm2.h>

const char *icu_plugin_version = DOVECOT_VERSION;

static char *icu_path;
static void *icu_mmap_data;
static size_t icu_mmap_len;
static const UNormalizer2 *icu_norm2;

static void icu_normalizer_init(void)
{
	UErrorCode err = 0;
	const char *name;
	struct stat st;
	int fd;

	name = strrchr(icu_path, '/');
	if (name == NULL)
		name = icu_path;
	else
		name++;
	name = t_strcut(name, '.');

	fd = open(icu_path, O_RDONLY);
	if (fd == -1)
		i_fatal("icu: Can't open %s: %m", icu_path);
	if (fstat(fd, &st) < 0)
		i_fatal("fstat(%s) failed: %m", icu_path);
	icu_mmap_len = st.st_size;
	icu_mmap_data = mmap(NULL, icu_mmap_len, PROT_READ, MAP_SHARED, fd, 0);
	if (icu_mmap_data == MAP_FAILED)
		i_fatal("mmap(%s) failed: %m", icu_path);
	udata_setAppData(name, icu_mmap_data, &err);
	if (U_FAILURE(err))
		i_fatal("icu: udata_setAppData() failed: %s", u_errorName(err));
	icu_norm2 = unorm2_getInstance(name, name, UNORM2_COMPOSE, &err);
	if (U_FAILURE(err))
		i_fatal("icu: unorm2_getInstance(%s) failed: %s",
			icu_path, u_errorName(err));
}

static int icu_normalizer(const void *_input, size_t size, buffer_t *output)
{
	const char *input = _input;
	int ret = 0;

	if (icu_norm2 == NULL)
		icu_normalizer_init();

	if (size > 0) T_BEGIN {
		UChar *icu_input = t_malloc(sizeof(UChar) * size);
		/* normalization shouldn't increase the size?
		   we'll allow some extra anyway.. */
		int32_t norm_output_max_len = size*2;
		UChar *norm_output = t_malloc(sizeof(UChar) * norm_output_max_len);
		int32_t icu_input_len, norm_len, buf_len, buf_len2;
		int32_t error_count = 0;
		char *buf;
		UErrorCode err = 0;

		/* first convert input to UChars */
		u_strFromUTF8WithSub(icu_input, size, &icu_input_len,
				     input, size, 0xfffd, &error_count, &err);
		if (U_FAILURE(err)) {
			i_fatal("icu: u_strFromUTF8WithSub(%s) failed: %s",
				icu_path, u_errorName(err));
		}
		/* then normalize it */
		norm_len = unorm2_normalize(icu_norm2, icu_input, icu_input_len,
					    norm_output, norm_output_max_len,
					    &err);
		if (U_FAILURE(err)) {
			i_fatal("icu: unorm2_normalize(%s) failed: %s",
				icu_path, u_errorName(err));
		}
		/* convert back to UTF-8. */
		u_strToUTF8(NULL, 0, &buf_len,
			    norm_output, norm_len, &err);
		err = 0;

		buf = buffer_append_space_unsafe(output, buf_len);
		u_strToUTF8(buf, buf_len, &buf_len2,
			    norm_output, norm_len, &err);
		if (U_FAILURE(err)) {
			i_fatal("icu: u_strToUTF8(%s) failed: %s",
				icu_path, u_errorName(err));
		}
		i_assert(buf_len == buf_len2);

		if (error_count > 0)
			ret = -1;
	} T_END;
	return ret;
}

static void icu_mail_user_created(struct mail_user *user)
{
	const char *value;

	value = mail_user_plugin_getenv(user, "icu_normalizer");
	if (value != NULL) {
		if (*value != '/') {
			/* relative path */
			value = t_strdup_printf(DOVECOT_PKGLIBDIR"/%s", value);
		}
		if (icu_path != NULL && strcmp(icu_path, value) != 0)
			i_fatal("Multiple normalizers not supported currently");
		icu_path = i_strdup(value);
		user->default_normalizer = icu_normalizer;
	}
}

static struct mail_storage_hooks icu_mail_storage_hooks = {
	.mail_user_created = icu_mail_user_created
};

void icu_plugin_init(struct module *module ATTR_UNUSED)
{
	mail_storage_hooks_add(module, &icu_mail_storage_hooks);
}

void icu_plugin_deinit(void)
{
	mail_storage_hooks_remove(&icu_mail_storage_hooks);
	i_free(icu_path);
}
